#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2025 Stéphane Caron and the qpsolvers contributors

"""Unit tests for jaxopt.OSQP."""

import unittest
import warnings

try:
    import jax.numpy as jnp

    from qpsolvers import solve_qp

    class TestKVXOPT(unittest.TestCase):
        """Test fixture for the KVXOPT solver."""

        def setUp(self):
            """Prepare test fixture."""
            warnings.simplefilter("ignore", category=UserWarning)

        def test_jax_array_input(self):
            """We can call ``solve_qp`` with jax.Array matrices."""
            M = jnp.array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
            P = M.T @ M  # this is a positive definite matrix
            q = jnp.array([3.0, 2.0, 3.0]) @ M
            G = jnp.array(
                [[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]]
            )
            h = jnp.array([3.0, 2.0, -2.0])
            A = jnp.array([1.0, 1.0, 1.0])
            b = jnp.array([1.0])
            x = solve_qp(P, q, G, h, A, b, solver="jaxopt_osqp")
            self.assertIsNotNone(x)

except ImportError as exn:  # in case the solver is not installed
    warnings.warn(f"Skipping jaxopt.OSQP tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for HiGHS."""

import unittest
import warnings

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.highs_ import highs_solve_qp

    class TestHiGHS(unittest.TestCase):
        """Test fixture for the HiGHS solver."""

        def setUp(self):
            """Prepare test fixture."""
            warnings.simplefilter("ignore", category=UserWarning)

        def test_highs_tolerances(self):
            problem = get_sd3310_problem()
            x = highs_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                time_limit=0.1,
                primal_feasibility_tolerance=1e-1,
                dual_feasibility_tolerance=1e-1,
            )
            self.assertIsNotNone(x)

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping HiGHS tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `PIQP`_.

.. _PIQP: https://github.com/PREDICT-EPFL/piqp

PIQP is a proximal interior-point quadratic programming solver for dense and
sparse problems. Its algorithm combines an infeasible interior-point method
with the proximal method of multipliers, and is designed to handle
ill-conditioned convex problems without the need for linear independence of
constraints. If you are using PIQP in a scientific work, consider citing the
corresponding paper [Schwan2023]_.
"""

import warnings
from typing import Optional, Union

import numpy as np
import piqp
import scipy.sparse as spa

from ..conversions import ensure_sparse_matrices
from ..exceptions import ParamError, ProblemError
from ..problem import Problem
from ..solution import Solution


def __select_backend(backend: Optional[str], use_csc: bool):
    """Select backend function for PIQP.

    Parameters
    ----------
    backend :
        PIQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    use_csc :
        If ``True``, use sparse matrices if the backend is not specified.

    Returns
    -------
    :
        Backend solve function.

    Raises
    ------
    ParamError
        If the required backend is not a valid PIQP backend.
    """
    if backend is None:
        return piqp.SparseSolver() if use_csc else piqp.DenseSolver()
    if backend == "dense":
        return piqp.DenseSolver()
    if backend == "sparse":
        return piqp.SparseSolver()
    raise ParamError(f'Unknown PIQP backend "{backend}')


def piqp_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    backend: Optional[str] = None,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using PIQP.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector (not used).
    backend :
        PIQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP returned by the solver.

    Notes
    -----
    All other keyword arguments are forwarded as options to PIQP. For
    instance, you can call ``piqp_solve_qp(P, q, G, h, eps_abs=1e-6)``.
    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - ``rho_init``
         - Initial value for the primal proximal penalty parameter rho.
       * - ``delta_init``
         - Initial value for the augmented lagrangian penalty parameter delta.
       * - ``eps_abs``
         - Absolute tolerance.
       * - ``eps_rel``
         - Relative tolerance.
       * - ``check_duality_gap``
         - Check terminal criterion on duality gap.
       * - ``eps_duality_gap_abs``
         - Absolute tolerance on duality gap.
       * - ``eps_duality_gap_rel``
         - Relative tolerance on duality gap.
       * - ``reg_lower_limit``
         - Lower limit for regularization.
       * - ``reg_finetune_lower_limit``
         - Fine tune lower limit regularization.
       * - ``reg_finetune_primal_update_threshold``
         - Threshold of number of no primal updates to transition to fine
           tune mode.
       * - ``reg_finetune_dual_update_threshold``
         - Threshold of number of no dual updates to transition to fine
           tune mode.
       * - ``max_iter``
         - Maximum number of iterations.
       * - ``max_factor_retires``
         - Maximum number of factorization retires before failure.
       * - ``preconditioner_scale_cost``
         - 	Scale cost in Ruiz preconditioner.
       * - ``preconditioner_iter``
         - Maximum of preconditioner iterations.
       * - ``tau``
         - Maximum interior point step length.
       * - ``iterative_refinement_always_enabled``
         - Always run iterative refinement and not only on factorization
           failure.
       * - ``iterative_refinement_eps_abs``
         - Iterative refinement absolute tolerance.
       * - ``iterative_refinement_eps_rel``
         - Iterative refinement relative tolerance.
       * - ``iterative_refinement_max_iter``
         - Maximum number of iterations for iterative refinement.
       * - ``iterative_refinement_min_improvement_rate``
         - Minimum improvement rate for iterative refinement.
       * - ``iterative_refinement_static_regularization_eps``
         - Static regularization for KKT system for iterative refinement.
       * - ``iterative_refinement_static_regularization_rel``
         - Static regularization w.r.t. the maximum abs diagonal term of
           KKT system.
       * - ``verbose``
         - Verbose printing.
       * - ``compute_timings``
         - Measure timing information internally.

    This list is not exhaustive. Check out the `solver documentation
    <https://predict-epfl.github.io/piqp/interfaces/settings>`__ for details.
    """
    P, q, G, h, A, b, lb, ub = problem.unpack()
    n: int = q.shape[0]

    if initvals is not None and verbose:
        warnings.warn("warm-start values are ignored by PIQP")

    if G is None and h is not None:
        raise ProblemError(
            "Inconsistent inequalities: G is not set but h is set"
        )
    if G is not None and h is None:
        raise ProblemError("Inconsistent inequalities: G is set but h is None")
    if A is None and b is not None:
        raise ProblemError(
            "Inconsistent inequalities: A is not set but b is set"
        )
    if A is not None and b is None:
        raise ProblemError("Inconsistent inequalities: A is set but b is None")
    # PIQP does not support A, b, G, and h to be None.
    use_csc: bool = (
        not isinstance(P, np.ndarray)
        or (G is not None and not isinstance(G, np.ndarray))
        or (A is not None and not isinstance(A, np.ndarray))
    )
    G_piqp = (
        G
        if G is not None
        else spa.csc_matrix(np.zeros((0, n))) if use_csc else np.zeros((0, n))
    )
    A_piqp = (
        A
        if A is not None
        else spa.csc_matrix(np.zeros((0, n))) if use_csc else np.zeros((0, n))
    )
    h_piqp = np.zeros((0,)) if h is None else h
    b_piqp = np.zeros((0,)) if b is None else b
    if use_csc:
        P, G_piqp, A_piqp = ensure_sparse_matrices("piqp", P, G_piqp, A_piqp)

    solver = __select_backend(backend, use_csc)
    solver.settings.verbose = verbose
    for key, value in kwargs.items():
        try:
            setattr(solver.settings, key, value)
        except AttributeError:
            if verbose:
                warnings.warn(
                    f"Received an undefined solver setting {key}\
                    with value {value}"
                )
    solver.setup(P, q, A_piqp, b_piqp, G_piqp, h_piqp, lb, ub)
    status = solver.solve()
    success_status = piqp.PIQP_SOLVED

    solution = Solution(problem)
    solution.extras = {"info": solver.result.info}
    solution.found = status == success_status
    solution.x = solver.result.x
    if A is None:
        solution.y = np.empty((0,))
    else:
        solution.y = solver.result.y
    if G is None:
        solution.z = np.empty((0,))
    else:
        solution.z = solver.result.z
    if lb is not None or ub is not None:
        solution.z_box = solver.result.z_ub - solver.result.z_lb
    else:
        solution.z_box = np.empty((0,))
    return solution


def piqp_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: Union[np.ndarray, spa.csc_matrix],
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    lb: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    ub: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    backend: Optional[str] = None,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using PIQP.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
        \underset{\mbox{minimize}}{x} &
            \frac{1}{2} x^T P x + q^T x \\
        \mbox{subject to}
            & G x \leq h                \\
            & A x = b                   \\
            & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `PIQP
    <https://github.com/PREDICT-EPFL/piqp>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    backend :
        PIQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    verbose :
        Set to `True` to print out extra information.
    initvals :
        Warm-start guess vector. Not used.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = piqp_solve_problem(
        problem, initvals, verbose, backend, **kwargs
    )
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `KVXOPT <https://github.com/sanurielf/kvxopt>`__.

KVXOPT is a fork from CVXOPT including more SuiteSparse functions
and KLU sparse matrix solver. As CVXOPT, it is a free, open-source
interior-point solver.
"""

import warnings
from typing import Dict, Optional, Union

import kvxopt
import numpy as np
import scipy.sparse as spa
from kvxopt.solvers import qp

from ..conversions import linear_from_box_inequalities, split_dual_linear_box
from ..exceptions import ProblemError, SolverError
from ..problem import Problem
from ..solution import Solution

kvxopt.solvers.options["show_progress"] = False  # disable default verbosity


def __to_cvxopt(
    M: Union[np.ndarray, spa.csc_matrix],
) -> Union[kvxopt.matrix, kvxopt.spmatrix]:
    """Convert matrix to CVXOPT format.

    Parameters
    ----------
    M :
        Matrix in NumPy or CVXOPT format.

    Returns
    -------
    :
        Matrix in CVXOPT format.
    """
    if isinstance(M, np.ndarray):
        __infty__ = 1e10  # 1e20 tends to yield division-by-zero errors
        M_noinf = np.nan_to_num(M, posinf=__infty__, neginf=-__infty__)
        return kvxopt.matrix(M_noinf)
    coo = M.tocoo()
    return kvxopt.spmatrix(
        coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=M.shape
    )


def kvxopt_solve_problem(
    problem: Problem,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    r"""Solve a quadratic program using KVXOPT.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    solver :
        Set to 'mosek' to run MOSEK rather than KVXOPT.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Raises
    ------
    ProblemError
        If the KVXOPT rank assumption is not satisfied.

    SolverError
        If KVXOPT failed with an error.

    Note
    ----
    .. _KVXOPT rank assumptions:

    **Rank assumptions:** KVXOPT requires the QP matrices to satisfy the

    .. math::

        \begin{split}\begin{array}{cc}
        \mathrm{rank}(A) = p
        &
        \mathrm{rank}([P\ A^T\ G^T]) = n
        \end{array}\end{split}

    where :math:`p` is the number of rows of :math:`A` and :math:`n` is the
    number of optimization variables. See the "Rank assumptions" paragraph in
    the report `The CVXOPT linear and quadratic cone program solvers
    <http://www.ee.ucla.edu/~vandenbe/publications/coneprog.pdf>`_ for details.

    Notes
    -----
    KVXOPT only considers the lower entries of :math:`P`, therefore it will use
    a different cost than the one intended if a non-symmetric matrix is
    provided.

    Keyword arguments are forwarded as options to KVXOPT. For instance, we can
    call ``kvxopt_solve_qp(P, q, G, h, u, abstol=1e-4, reltol=1e-4)``. KVXOPT
    options include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``abstol``
         - Absolute tolerance on the duality gap.
       * - ``feastol``
         - Tolerance on feasibility conditions, that is, on the primal
           residual.
       * - ``maxiters``
         - Maximum number of iterations.
       * - ``refinement``
         - Number of iterative refinement steps when solving KKT equations
       * - ``reltol``
         - Relative tolerance on the duality gap.

    Check out `Algorithm Parameters
    <https://cvxopt.org/userguide/coneprog.html#algorithm-parameters>`_ section
    of the solver documentation for details and default values of all solver
    parameters. See also [Caron2022]_ for a primer on the duality gap, primal
    and dual residuals.
    """
    P, q, G, h, A, b, lb, ub = problem.unpack()
    if lb is not None or ub is not None:
        G, h = linear_from_box_inequalities(
            G, h, lb, ub, use_sparse=problem.has_sparse
        )

    args = [__to_cvxopt(P), __to_cvxopt(q)]
    constraints = {"G": None, "h": None, "A": None, "b": None}
    if G is not None and h is not None:
        constraints["G"] = __to_cvxopt(G)
        constraints["h"] = __to_cvxopt(h)
    if A is not None and b is not None:
        constraints["A"] = __to_cvxopt(A)
        constraints["b"] = __to_cvxopt(b)
    initvals_dict: Optional[Dict[str, kvxopt.matrix]] = None
    if initvals is not None:
        if "mosek" in kwargs:
            warnings.warn("MOSEK: warm-start values are ignored")
        initvals_dict = {"x": __to_cvxopt(initvals)}
    kwargs["show_progress"] = verbose

    try:
        res = qp(
            *args,
            solver=solver,
            initvals=initvals_dict,
            options=kwargs,
            **constraints,
        )
    except ValueError as exception:
        error = str(exception)
        if "Rank(A)" in error:
            raise ProblemError(error) from exception
        raise SolverError(error) from exception

    solution = Solution(problem)
    solution.extras = res
    solution.found = "optimal" in res["status"]
    solution.x = np.array(res["x"]).flatten()
    solution.y = (
        np.array(res["y"]).flatten() if b is not None else np.empty((0,))
    )
    if h is not None and res["z"] is not None:
        z_cvxopt = np.array(res["z"]).flatten()
        if z_cvxopt.size == h.size:
            z, z_box = split_dual_linear_box(z_cvxopt, lb, ub)
            solution.z = z
            solution.z_box = z_box
    else:  # h is None
        solution.z = np.empty((0,))
        solution.z_box = np.empty((0,))
    solution.obj = res["primal objective"]
    return solution


def kvxopt_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using KVXOPT.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `KVXOPT <https://github.com/sanurielf/kvxopt>`__.

    Parameters
    ----------
    P :
        Symmetric cost matrix. Together with :math:`A` and :math:`G`, it should
        satisfy :math:`\mathrm{rank}([P\ A^T\ G^T]) = n`, see the rank
        assumptions below.
    q :
        Cost vector.
    G :
        Linear inequality matrix. Together with :math:`P` and :math:`A`, it
        should satisfy :math:`\mathrm{rank}([P\ A^T\ G^T]) = n`, see the
        rank assumptions below.
    h :
        Linear inequality vector.
    A :
        Linear equality constraint matrix. It needs to be full row rank, and
        together with :math:`P` and :math:`G` satisfy
        :math:`\mathrm{rank}([P\ A^T\ G^T]) = n`. See the rank assumptions
        below.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    solver :
        Set to 'mosek' to run MOSEK rather than KVXOPT.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.

    Raises
    ------
    ProblemError
        If the KVXOPT rank assumption is not satisfied.

    SolverError
        If KVXOPT failed with an error.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = kvxopt_solve_problem(
        problem, solver, initvals, verbose, **kwargs
    )
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2024 Lev Kozlov

"""Unit tests for qpax."""

import unittest
import warnings

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.qpax_ import qpax_solve_qp

    class TestQpax(unittest.TestCase):
        """Tests specific to qpax."""

        def test_problem(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, lb, ub = problem.unpack()
            self.assertIsNotNone(qpax_solve_qp(P, q, G, h, A, b, lb, ub))


except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping qpax tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Convert quadratic programs to second-order cone programs."""

from typing import Any, Dict, Optional, Tuple

from numpy import hstack, ndarray, sqrt, vstack, zeros
from numpy.linalg import LinAlgError, cholesky
from scipy.sparse import csc_matrix

from ..exceptions import ProblemError


def socp_from_qp(
    P: ndarray, q: ndarray, G: Optional[ndarray], h: Optional[ndarray]
) -> Tuple[ndarray, ndarray, ndarray, Dict[str, Any]]:
    r"""Convert a quadratic program to a second-order cone program.

    The quadratic program is defined by:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h
        \end{array}\end{split}

    The equivalent second-order cone program is:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                c^T_s y \\
            \mbox{subject to}
                & G_s y \leq_{\cal K} h_s
        \end{array}\end{split}

    This function is adapted from ``ecosqp.m`` in the `ecos-matlab
    <https://github.com/embotech/ecos-matlab/>`_ repository. See the
    documentation in that script for details on this reformulation.

    Parameters
    ----------
    P :
        Primal quadratic cost matrix.
    q :
        Primal quadratic cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.

    Returns
    -------
    c_socp : array
        SOCP cost vector.
    G_socp : array
        SOCP inequality matrix.
    h_socp : array
        SOCP inequality vector.
    dims : dict
        Dimension dictionary used by SOCP solvers, where ``dims["l"]`` is the
        number of inequality constraints.

    Raises
    ------
    ValueError :
        If the cost matrix is not positive definite.
    """
    n = P.shape[1]  # dimension of QP variable
    c_socp = hstack([zeros(n), 1])  # new SOCP variable stacked as [x, t]
    try:
        L = cholesky(P)
    except LinAlgError as e:
        error = str(e)
        if "not positive definite" in error:
            raise ProblemError("matrix P is not positive definite") from e
        raise e  # other linear algebraic error

    scale = 1.0 / sqrt(2)
    G_quad = vstack(
        [
            scale * hstack([q, -1.0]),
            hstack([-L.T, zeros((L.shape[0], 1))]),
            scale * hstack([-q, +1.0]),
        ]
    )
    h_quad = hstack([scale, zeros(L.shape[0]), scale])

    dims: Dict[str, Any] = {"q": [L.shape[0] + 2]}
    if G is not None and h is not None:
        G_socp = vstack([hstack([G, zeros((G.shape[0], 1))]), G_quad])
        h_socp = hstack([h, h_quad])
        dims["l"] = G.shape[0]
    else:  # no linear inequality constraint
        G_socp = G_quad
        h_socp = h_quad
        dims["l"] = 0

    G_socp = csc_matrix(G_socp)
    return c_socp, G_socp, h_socp, dims



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for qpOASES."""

import unittest
import warnings

import numpy as np
import scipy.sparse as spa

from qpsolvers import ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.qpoases_ import qpoases_solve_qp

    class TestQpOASES(unittest.TestCase):
        """Test fixture specific to the qpOASES solver."""

        def test_initvals(self):
            """Call the solver with a warm-start guess."""
            problem = get_sd3310_problem()
            qpoases_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                initvals=problem.q,
            )

        def test_params(self):
            """Call the solver with a time limit and other parameters."""
            problem = get_sd3310_problem()
            qpoases_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                time_limit=0.1,
                terminationTolerance=1e-7,
            )
            qpoases_solve_qp(
                problem.P,
                problem.q,
                time_limit=0.1,
                terminationTolerance=1e-7,
            )

        def test_unfeasible(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, _, _ = problem.unpack()
            custom_lb = np.ones(q.shape)
            custom_ub = -np.ones(q.shape)
            x = qpoases_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                custom_lb,
                custom_ub,
            )
            self.assertIsNone(x)

        def test_not_sparse(self):
            """Raise a ProblemError on sparse problems."""
            problem = get_sd3310_problem()
            problem.P = spa.csc_matrix(problem.P)
            with self.assertRaises(ProblemError):
                qpoases_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                )

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping qpOASES tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors
# Copyright 2021 Dustin Kenefake

"""Solver interface for `Gurobi <https://www.gurobi.com/>`__.

The Gurobi Optimizer suite ships several solvers for mathematical programming,
including problems that have linear constraints, bound constraints, integrality
constraints, cone constraints, or quadratic constraints. It targets modern CPU
architectures and multi-core processors,

See the :ref:`installation page <gurobi-install>` for additional instructions
on installing this solver.
"""

import warnings
from typing import Optional, Union

import gurobipy
import numpy as np
import scipy.sparse as spa
from gurobipy import GRB

from ..problem import Problem
from ..solution import Solution


def gurobi_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using Gurobi.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution returned by the solver.

    Notes
    -----
    Keyword arguments are forwarded to Gurobi as parameters. For instance, we
    can call ``gurobi_solve_qp(P, q, G, h, u, FeasibilityTol=1e-8,
    OptimalityTol=1e-8)``. Gurobi settings include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``FeasibilityTol``
         - Primal feasibility tolerance.
       * - ``OptimalityTol``
         - Dual feasibility tolerance.
       * - ``PSDTol``
         - Positive semi-definite tolerance.
       * - ``TimeLimit``
         - Run time limit in seconds, 0 to disable.

    Check out the `Parameter Descriptions
    <https://www.gurobi.com/documentation/9.5/refman/parameter_descriptions.html>`_
    documentation for all available Gurobi parameters.

    Lower values for primal or dual tolerances yield more precise solutions at
    the cost of computation time. See *e.g.* [Caron2022]_ for a primer of
    solver tolerances.
    """
    if initvals is not None:
        warnings.warn("warm-start values are ignored by this wrapper")

    model = gurobipy.Model()
    if not verbose:
        model.setParam(GRB.Param.OutputFlag, 0)
    for param, value in kwargs.items():
        model.setParam(param, value)

    P, q, G, h, A, b, lb, ub = problem.unpack()
    num_vars = P.shape[0]
    identity = spa.eye(num_vars)
    x = model.addMVar(
        num_vars, lb=-GRB.INFINITY, ub=GRB.INFINITY, vtype=GRB.CONTINUOUS
    )
    ineq_constr, eq_constr, lb_constr, ub_constr = None, None, None, None
    if G is not None:
        ineq_constr = model.addMConstr(G, x, GRB.LESS_EQUAL, h)
    if A is not None:
        eq_constr = model.addMConstr(A, x, GRB.EQUAL, b)
    if lb is not None:
        lb_constr = model.addMConstr(identity, x, GRB.GREATER_EQUAL, lb)
    if ub is not None:
        ub_constr = model.addMConstr(identity, x, GRB.LESS_EQUAL, ub)
    objective = 0.5 * (x @ P @ x) + q @ x
    model.setObjective(objective, sense=GRB.MINIMIZE)
    model.optimize()

    solution = Solution(problem)
    solution.extras["status"] = model.status
    solution.found = model.status in (GRB.OPTIMAL, GRB.SUBOPTIMAL)
    if solution.found:
        solution.x = x.X
        __retrieve_dual(solution, ineq_constr, eq_constr, lb_constr, ub_constr)
    return solution


def __retrieve_dual(
    solution: Solution,
    ineq_constr: Optional[gurobipy.MConstr],
    eq_constr: Optional[gurobipy.MConstr],
    lb_constr: Optional[gurobipy.MConstr],
    ub_constr: Optional[gurobipy.MConstr],
) -> None:
    solution.z = -ineq_constr.Pi if ineq_constr is not None else np.empty((0,))
    solution.y = -eq_constr.Pi if eq_constr is not None else np.empty((0,))
    if lb_constr is not None and ub_constr is not None:
        solution.z_box = -ub_constr.Pi - lb_constr.Pi
    elif ub_constr is not None:  # lb_constr is None
        solution.z_box = -ub_constr.Pi
    elif lb_constr is not None:  # ub_constr is None
        solution.z_box = -lb_constr.Pi
    else:  # lb_constr is None and ub_constr is None
        solution.z_box = np.empty((0,))


def gurobi_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using Gurobi.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `Gurobi <http://www.gurobi.com/>`__.

    Parameters
    ----------
    P :
        Primal quadratic cost matrix.
    q :
        Primal quadratic cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Notes
    -----
    Keyword arguments are forwarded to Gurobi as parameters. For instance, we
    can call ``gurobi_solve_qp(P, q, G, h, u, FeasibilityTol=1e-8,
    OptimalityTol=1e-8)``. Gurobi settings include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``FeasibilityTol``
         - Primal feasibility tolerance.
       * - ``OptimalityTol``
         - Dual feasibility tolerance.
       * - ``PSDTol``
         - Positive semi-definite tolerance.
       * - ``TimeLimit``
         - Run time limit in seconds, 0 to disable.

    Check out the `Parameter Descriptions
    <https://www.gurobi.com/documentation/9.5/refman/parameter_descriptions.html>`_
    documentation for all available Gurobi parameters.

    Lower values for primal or dual tolerances yield more precise solutions at
    the cost of computation time. See *e.g.* [Caron2022]_ for a primer of
    solver tolerances.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = gurobi_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `ECOS <https://github.com/embotech/ecos>`__.

ECOS is an interior-point solver for convex second-order cone programs (SOCPs).
designed specifically for embedded applications. ECOS is written in low
footprint, single-threaded, library-free ANSI-C and so runs on most embedded
platforms. For small problems, ECOS is faster than most existing SOCP solvers;
it is still competitive for medium-sized problems up to tens of thousands of
variables. If you are using ECOS in a scientific work, consider citing the
corresponding paper [Domahidi2013]_.
"""

import warnings
from typing import Optional, Union

import numpy as np
from ecos import solve
from scipy import sparse as spa

from ..conversions import (
    linear_from_box_inequalities,
    socp_from_qp,
    split_dual_linear_box,
)
from ..exceptions import ProblemError
from ..problem import Problem
from ..solution import Solution

__exit_flag_meaning__ = {
    0: "OPTIMAL",
    1: "PINF: found certificate of primal infeasibility",
    2: "DING: found certificate of dual infeasibility",
    10: "INACC_OFFSET: inaccurate results",
    11: "PINF_INACC: found inaccurate certificate of primal infeasibility",
    12: "DING_INACC: found inaccurate certificate of dual infeasibility",
    -1: "MAXIT: maximum number of iterations reached",
    -2: "NUMERICS: search direction is unreliable",
    -3: "OUTCONE: primal or dual variables got outside of cone",
    -4: "SIGINT: solver interrupted",
    -7: "FATAL: unknown solver problem",
}


def ecos_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using ECOS.

    Parameters
    ----------
    P :
        Primal quadratic cost matrix.
    q :
        Primal quadratic cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Raises
    ------
    ProblemError :
        If inequality constraints contain infinite values that the solver
        doesn't handle.

    ValueError :
        If the cost matrix is not positive definite.

    Notes
    -----
    All other keyword arguments are forwarded as options to the ECOS solver.
    For instance, you can call ``qpswift_solve_qp(P, q, G, h, abstol=1e-5)``.

    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - ``feastol``
         -  Tolerance on the primal and dual residual.
       * - ``abstol``
         -  Absolute tolerance on the duality gap.
       * - ``reltol``
         -  Relative tolerance on the duality gap.
       * - ``feastol_inacc``
         -  Tolerance on the primal and dual residual if reduced precisions.
       * - ``abstol_inacc``
         - Absolute tolerance on the duality gap if reduced precision.
       * - ``reltolL_inacc``
         - Relative tolerance on the duality gap if reduced precision.
       * - ``max_iters``
         - Maximum numer of iterations.
       * - ``nitref``
         - Number of iterative refinement steps.

    See the `ECOS Python wrapper documentation
    <https://github.com/embotech/ecos-python#calling-ecos-from-python>`_ for
    more details. You can also check out [Caron2022]_ for a primer on
    primal-dual residuals or the duality gap.
    """
    if initvals is not None:
        warnings.warn("warm-start values are ignored by this wrapper")
    P, q, G, h, A, b, lb, ub = problem.unpack()
    if lb is not None or ub is not None:
        G, h = linear_from_box_inequalities(
            G, h, lb, ub, use_sparse=problem.has_sparse
        )
    kwargs.update({"verbose": verbose})
    c_socp, G_socp, h_socp, dims = socp_from_qp(P, q, G, h)
    if A is not None:
        A_socp = spa.hstack([A, spa.csc_matrix((A.shape[0], 1))], format="csc")
        result = solve(c_socp, G_socp, h_socp, dims, A_socp, b, **kwargs)
    else:
        result = solve(c_socp, G_socp, h_socp, dims, **kwargs)
    flag = result["info"]["exitFlag"]
    solution = Solution(problem)
    solution.extras = result["info"]
    solution.found = flag == 0
    if not solution.found:
        if h is not None and not np.isfinite(h).all():
            raise ProblemError(
                "ECOS does not handle infinite values in inequality vectors, "
                "try clipping them to a finite value suitable to your problem"
            )
        meaning = __exit_flag_meaning__.get(flag, "unknown exit flag")
        warnings.warn(f"ECOS returned exit flag {flag} ({meaning})")
    solution.x = result["x"][:-1]
    if A is not None:
        solution.y = result["y"]
    if G is not None:
        z_ecos = result["z"][: G.shape[0]]
        z, z_box = split_dual_linear_box(z_ecos, lb, ub)
        solution.z = z
        solution.z_box = z_box
    return solution


def ecos_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using ECOS.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b
        \end{array}\end{split}

    It is solved using `ECOS <https://github.com/embotech/ecos>`__.

    Parameters
    ----------
    P :
        Primal quadratic cost matrix.
    q :
        Primal quadratic cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Notes
    -----
    All other keyword arguments are forwarded as options to the ECOS solver.
    For instance, you can call ``ecos_solve_qp(P, q, G, h, abstol=1e-5)``.

    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - ``feastol``
         -  Tolerance on the primal and dual residual.
       * - ``abstol``
         -  Absolute tolerance on the duality gap.
       * - ``reltol``
         -  Relative tolerance on the duality gap.
       * - ``feastol_inacc``
         -  Tolerance on the primal and dual residual if reduced precisions.
       * - ``abstol_inacc``
         - Absolute tolerance on the duality gap if reduced precision.
       * - ``reltolL_inacc``
         - Relative tolerance on the duality gap if reduced precision.
       * - ``max_iters``
         - Maximum numer of iterations.
       * - ``nitref``
         - Number of iterative refinement steps.

    See the `ECOS Python wrapper documentation
    <https://github.com/embotech/ecos-python#calling-ecos-from-python>`_ for
    more details. You can also check out [Caron2022]_ for a primer on
    primal-dual residuals or the duality gap.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = ecos_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2025 Inria

"""Warnings from qpsolvers."""


class QPWarning(UserWarning):
    """Base class for qpsolvers warnings."""


class SparseConversionWarning(QPWarning):
    """Warning issued when converting NumPy arrays to SciPy sparse matrices."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for Clarabel."""

import unittest
import warnings

from qpsolvers.problems import get_qpsut01

try:
    import clarabel

    from qpsolvers.solvers.clarabel_ import clarabel_solve_problem

    class TestClarabel(unittest.TestCase):
        """Test fixture for the Clarabel.rs solver."""

        def test_time_limit(self):
            """Call Clarabel.rs with an infeasibly low time limit."""
            problem, ref_solution = get_qpsut01()
            solution = clarabel_solve_problem(problem, time_limit=1e-10)
            status = solution.extras["status"]
            self.assertEqual(status, clarabel.SolverStatus.MaxTime)
            # See https://github.com/oxfordcontrol/Clarabel.rs/issues/10
            self.assertFalse(status != clarabel.SolverStatus.MaxTime)

        def test_status(self):
            """Check that result status is consistent with its string repr.

            Context: https://github.com/oxfordcontrol/Clarabel.rs/issues/10
            """
            problem, _ = get_qpsut01()
            solution = clarabel_solve_problem(problem)
            status = solution.extras["status"]
            check_1 = str(status) != "Solved"
            check_2 = status != clarabel.SolverStatus.Solved
            self.assertEqual(check_1, check_2)

except ImportError as exn:  # in case the solver is not installed
    warnings.warn(f"Skipping Clarabel.rs tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Convert stacked dual multipliers into linear and box multipliers."""

from typing import Optional, Tuple

import numpy as np


def split_dual_linear_box(
    z_stacked: np.ndarray,
    lb: Optional[np.ndarray],
    ub: Optional[np.ndarray],
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    """Separate linear and box multipliers from a stacked dual vector.

    This function assumes linear and box inequalities were combined using
    :func:`qpsolvers.conversions.linear_from_box_inequalities`.

    Parameters
    ----------
    z_stacked :
        Stacked vector of dual multipliers.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.

    Returns
    -------
    :
        Pair :code:`z, z_box` of linear and box multipliers. Both can be empty
        arrays if there is no corresponding constraint.
    """
    z = np.empty((0,), dtype=z_stacked.dtype)
    z_box = np.empty((0,), dtype=z_stacked.dtype)
    if lb is not None and ub is not None:
        n_lb = lb.shape[0]
        n_ub = ub.shape[0]
        n_box = n_lb + n_ub
        z_box = z_stacked[-n_ub:] - z_stacked[-n_box:-n_ub]
        z = z_stacked[:-n_box]
    elif ub is not None:  # lb is None
        n_ub = ub.shape[0]
        z_box = z_stacked[-n_ub:]
        z = z_stacked[:-n_ub]
    elif lb is not None:  # ub is None
        n_lb = lb.shape[0]
        z_box = -z_stacked[-n_lb:]
        z = z_stacked[:-n_lb]
    else:  # lb is None and ub is None
        z = z_stacked
    return z, z_box



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test a random available QP solver on a small least-squares problem."""

import random
from time import perf_counter

import numpy as np

import qpsolvers
from qpsolvers import print_matrix_vector, solve_ls

R = np.array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
s = np.array([3.0, 2.0, 3.0])
G = np.array([[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]])
h = np.array([3.0, 2.0, -2.0])

x_sol = np.array([0.1299734765610818, -0.0649867382805409, 1.7400530468778364])

if __name__ == "__main__":
    start_time = perf_counter()
    solver = random.choice(qpsolvers.available_solvers)
    x = solve_ls(R, s, G, h, solver=solver, verbose=False)
    end_time = perf_counter()

    print("")
    print("    min. || R * x - s ||^2")
    print("    s.t. G * x <= h")
    print("")
    print_matrix_vector(R, "R", s, "s")
    print("")
    print_matrix_vector(G, "G", h, "h")
    print("")
    print(f"Solution: x = {x}")
    print(f"It should be close to x* = {x_sol}")
    print(f"Found in {1e6 * (end_time - start_time):.0f} [us] with {solver}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test all available QP solvers on a sparse quadratic program."""

from os.path import basename

import numpy as np
import scipy.sparse
from IPython import get_ipython
from numpy.linalg import norm
from scipy.sparse import csc_matrix

from qpsolvers import dense_solvers, solve_qp, sparse_solvers

n = 500
M = scipy.sparse.lil_matrix(scipy.sparse.eye(n))
for i in range(1, n - 1):
    M[i, i + 1] = -1
    M[i, i - 1] = 1
P = csc_matrix(M.dot(M.transpose()))
q = -np.ones((n,))
G = csc_matrix(-scipy.sparse.eye(n))
h = -2 * np.ones((n,))
P_array = np.array(P.todense())
G_array = np.array(G.todense())


def check_same_solutions(tol=0.05):
    sol0 = solve_qp(P, q, G, h, solver=sparse_solvers[0])
    for solver in sparse_solvers:
        sol = solve_qp(P, q, G, h, solver=solver)
        relvar = norm(sol - sol0) / norm(sol0)
        assert (
            relvar < tol
        ), f"{solver}'s solution offset by {100.0 * relvar:.1f}%"
    for solver in dense_solvers:
        sol = solve_qp(P_array, q, G_array, h, solver=solver)
        relvar = norm(sol - sol0) / norm(sol0)
        assert (
            relvar < tol
        ), f"{solver}'s solution offset by {100.0 * relvar:.1f}%"


def time_dense_solvers():
    instructions = {
        solver: f"u = solve_qp(P_array, q, G_array, h, solver='{solver}')"
        for solver in dense_solvers
    }
    print("\nDense solvers\n-------------")
    for solver, instr in instructions.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)


def time_sparse_solvers():
    instructions = {
        solver: f"u = solve_qp(P, q, G, h, solver='{solver}')"
        for solver in sparse_solvers
    }
    print("\nSparse solvers\n--------------")
    for solver, instr in instructions.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)


if __name__ == "__main__":
    if get_ipython() is None:
        print(
            "This example should be run with IPython:\n\n"
            f"\tipython -i {basename(__file__)}\n"
        )
        exit()

    benchmark = "https://github.com/qpsolvers/qpbenchmark"
    print("\nTesting all QP solvers on one given sparse quadratic program")
    print(f"For a proper benchmark, check out {benchmark}")

    check_same_solutions()
    time_dense_solvers()
    time_sparse_solvers()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test all available QP solvers on a dense quadratic program."""

from os.path import basename

from IPython import get_ipython
from numpy import array, dot
from qpsolvers import dense_solvers, solve_qp, sparse_solvers
from scipy.sparse import csc_matrix

M = array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
P = dot(M.T, M)
q = dot(array([3.0, 2.0, 3.0]), M)
G = array([[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]])
h = array([3.0, 2.0, -2.0])
P_csc = csc_matrix(P)
G_csc = csc_matrix(G)


if __name__ == "__main__":
    if get_ipython() is None:
        print(
            "This example should be run with IPython:\n\n"
            f"\tipython -i {basename(__file__)}\n"
        )
        exit()

    dense_instr = {
        solver: f"u = solve_qp(P, q, G, h, solver='{solver}')"
        for solver in dense_solvers
    }
    sparse_instr = {
        solver: f"u = solve_qp(P_csc, q, G_csc, h, solver='{solver}')"
        for solver in sparse_solvers
    }

    benchmark = "https://github.com/qpsolvers/qpbenchmark"
    print("\nTesting all QP solvers on one given dense quadratic program")
    print(f"For a proper benchmark, check out {benchmark}")

    sol0 = solve_qp(P, q, G, h, solver=dense_solvers[0])
    abstol = 2e-4  # tolerance on absolute solution error
    for solver in dense_solvers:
        sol = solve_qp(P, q, G, h, solver=solver)
    for solver in sparse_solvers:
        sol = solve_qp(P_csc, q, G_csc, h, solver=solver)

    print("\nDense solvers\n-------------")
    for solver, instr in dense_instr.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)

    print("\nSparse solvers\n--------------")
    for solver, instr in sparse_instr.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `ProxQP`_.

.. _ProxQP: https://github.com/Simple-Robotics/proxsuite#proxqp

ProxQP is a primal-dual augmented Lagrangian method with proximal heuristics.
It converges to the solution of feasible problems, or to the solution to the
closest feasible one if the input problem is unfeasible. ProxQP is part of the
ProxSuite collection of open-source solvers. If you use ProxQP in a scientific
work, consider citing the corresponding paper [Bambade2022]_.
"""

from typing import Optional, Union

import numpy as np
import scipy.sparse as spa
from proxsuite import proxqp

from ..conversions import combine_linear_box_inequalities
from ..exceptions import ParamError
from ..problem import Problem
from ..solution import Solution


def __select_backend(backend: Optional[str], use_csc: bool):
    """Select backend function for ProxQP.

    Parameters
    ----------
    backend :
        ProxQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    use_csc :
        If ``True``, use sparse matrices if the backend is not specified.

    Returns
    -------
    :
        Backend solve function.

    Raises
    ------
    ParamError
        If the required backend is not a valid ProxQP backend.
    """
    if backend is None:
        return proxqp.sparse.solve if use_csc else proxqp.dense.solve
    if backend == "dense":
        return proxqp.dense.solve
    if backend == "sparse":
        return proxqp.sparse.solve
    raise ParamError(f'Unknown ProxQP backend "{backend}')


def proxqp_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    backend: Optional[str] = None,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using ProxQP.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector.
    backend :
        ProxQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP returned by the solver.

    Raises
    ------
    ParamError
        If a warm-start value is given both in `initvals` and the `x` keyword
        argument.

    Notes
    -----
    All other keyword arguments are forwarded as options to ProxQP. For
    instance, you can call ``proxqp_solve_qp(P, q, G, h, eps_abs=1e-6)``.
    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - ``x``
         - Warm start value for the primal variable.
       * - ``y``
         - Warm start value for the dual Lagrange multiplier for equality
           constraints.
       * - ``z``
         - Warm start value for the dual Lagrange multiplier for inequality
           constraints.
       * - ``eps_abs``
         - Asbolute stopping criterion of the solver (default: 1e-3, note that
           this is a laxer default than other solvers). See *e.g.*
           [Caron2022]_ for an overview of solver tolerances.
       * - ``eps_rel``
         - Relative stopping criterion of the solver. See *e.g.* [Caron2022]_
           for an overview of solver tolerances.
       * - ``check_duality_gap``
         - If set to true (false by default), ProxQP will include the duality
           gap in absolute and relative stopping criteria.
       * - ``mu_eq``
         - Proximal step size wrt equality constraints multiplier.
       * - ``mu_in``
         - Proximal step size wrt inequality constraints multiplier.
       * - ``rho``
         - Proximal step size wrt primal variable.
       * - ``compute_preconditioner``
         - If ``True`` (default), the preconditioner will be derived.
       * - ``compute_timings``
         - If ``True`` (default), timings will be computed by the solver (setup
           time, solving time, and run time = setup time + solving time).
       * - ``max_iter``
         - Maximal number of authorized outer iterations.
       * - ``initial_guess``
         - Sets the initial guess option for initilizing x, y and z.

    This list is not exhaustive. Check out the `solver documentation
    <https://simple-robotics.github.io/proxsuite/>`__ for details.
    """
    if initvals is not None:
        if "x" in kwargs:
            raise ParamError(
                "Warm-start value specified in both `initvals` and `x` kwargs"
            )
        kwargs["x"] = initvals
    P, q, G, h, A, b, lb, ub = problem.unpack()
    n: int = q.shape[0]
    use_csc: bool = (
        not isinstance(P, np.ndarray)
        or (G is not None and not isinstance(G, np.ndarray))
        or (A is not None and not isinstance(A, np.ndarray))
    )
    Cx, ux, lx = combine_linear_box_inequalities(G, h, lb, ub, n, use_csc)
    solve = __select_backend(backend, use_csc)
    result = solve(
        P,
        q,
        A,
        b,
        Cx,
        lx,
        ux,
        verbose=verbose,
        **kwargs,
    )
    solution = Solution(problem)
    solution.extras = {"info": result.info}
    solution.found = result.info.status == proxqp.QPSolverOutput.PROXQP_SOLVED
    solution.x = result.x
    solution.y = result.y
    if lb is not None or ub is not None:
        solution.z = result.z[:-n]
        solution.z_box = result.z[-n:]
    else:  # lb is None and ub is None
        solution.z = result.z
    return solution


def proxqp_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: Union[np.ndarray, spa.csc_matrix],
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    lb: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    ub: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    backend: Optional[str] = None,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using ProxQP.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
        \underset{\mbox{minimize}}{x} &
            \frac{1}{2} x^T P x + q^T x \\
        \mbox{subject to}
            & G x \leq h                \\
            & A x = b                   \\
            & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `ProxQP
    <https://github.com/Simple-Robotics/proxsuite#proxqp>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector.
    backend :
        ProxQP backend to use in ``[None, "dense", "sparse"]``. If ``None``
        (default), the backend is selected based on the type of ``P``.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = proxqp_solve_problem(
        problem, initvals, verbose, backend, **kwargs
    )
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test all available QP solvers on random quadratic programs."""

import sys

try:
    from IPython import get_ipython
except ImportError:
    print("This example requires IPython, try installing ipython3")
    sys.exit(-1)

from os.path import basename
from timeit import timeit

from numpy import dot, linspace, ones, random
from qpsolvers import available_solvers, solve_qp
from scipy.linalg import toeplitz

nb_iter = 10
sizes = [10, 20, 50, 100, 200, 500, 1000, 2000]


def solve_random_qp(n, solver):
    M, b = random.random((n, n)), random.random(n)
    P, q = dot(M.T, M), dot(b, M)
    G = toeplitz(
        [1.0, 0.0, 0.0] + [0.0] * (n - 3), [1.0, 2.0, 3.0] + [0.0] * (n - 3)
    )
    h = ones(n)
    return solve_qp(P, q, G, h, solver=solver)


def plot_results(perfs):
    try:
        from pylab import (
            clf,
            get_cmap,
            grid,
            ion,
            legend,
            plot,
            xlabel,
            xscale,
            ylabel,
            yscale,
        )
    except ImportError:
        print("Cannot plot results, try installing python3-matplotlib")
        print("Results are stored in the global `perfs` dictionary")
        return

    cmap = get_cmap("tab10")
    colors = cmap(linspace(0, 1, len(available_solvers)))
    solver_color = {
        solver: colors[i] for i, solver in enumerate(available_solvers)
    }
    ion()
    clf()
    for solver in perfs:
        plot(sizes, perfs[solver], lw=2, color=solver_color[solver])
    grid(True)
    legend(list(perfs.keys()), loc="lower right")
    xscale("log")
    yscale("log")
    xlabel("Problem size $n$")
    ylabel("Time (s)")
    for solver in perfs:
        plot(sizes, perfs[solver], marker="o", color=solver_color[solver])


if __name__ == "__main__":
    if get_ipython() is None:
        print(
            "This example should be run with IPython:\n\n"
            f"\tipython -i {basename(__file__)}\n"
        )
        exit()
    perfs = {}

    benchmark = "https://github.com/qpsolvers/qpbenchmark"
    print("\nTesting all solvers on a given set of random QPs")
    print(f"For a proper benchmark, check out {benchmark}")

    for solver in available_solvers:
        try:
            perfs[solver] = []
            for size in sizes:
                print(f"Running {solver} on problem size {size}...")
                cum_time = timeit(
                    stmt=f"solve_random_qp({size}, '{solver}')",
                    setup="from __main__ import solve_random_qp",
                    number=nb_iter,
                )
                perfs[solver].append(cum_time / nb_iter)
        except Exception as e:
            print(f"Warning: {str(e)}")
            if solver in perfs:
                del perfs[solver]
    plot_results(perfs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2023 Inria

"""Solve an unconstrained quadratic program."""

import numpy as np
from scipy.sparse.linalg import lsqr

from .exceptions import ProblemError
from .problem import Problem
from .solution import Solution


def solve_unconstrained(problem: Problem) -> Solution:
    """Solve an unconstrained quadratic program with SciPy's LSQR.

    Parameters
    ----------
    problem :
        Unconstrained quadratic program.

    Returns
    -------
    :
        Solution to the unconstrained QP, if it is bounded.

    Raises
    ------
    ValueError
        If the quadratic program is not unbounded below.
    """
    P, q, _, _, _, _, _, _ = problem.unpack()
    solution = Solution(problem)
    solution.x = lsqr(P, -q)[0]
    cost_check = np.linalg.norm(P @ solution.x + q)
    if cost_check > 1e-8:
        raise ProblemError(
            f"problem is unbounded below (cost_check={cost_check:.1e}), "
            "q has component in the nullspace of P"
        )
    solution.found = True
    return solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2023 Inria

"""Solver interface for `Clarabel.rs`_.

.. _Clarabel.rs: https://github.com/oxfordcontrol/Clarabel.rs

Clarabel.rs is a Rust implementation of an interior point numerical solver for
convex optimization problems using a novel homogeneous embedding. A paper
describing the Clarabel solver algorithm and implementation will be forthcoming
soon (retrieved: 2023-02-06). Until then, the authors ask that you cite its
documentation if you have found Clarabel.rs useful in your work.
"""

import warnings
from typing import Optional, Union

import clarabel
import numpy as np
import scipy.sparse as spa

from ..conversions import (
    ensure_sparse_matrices,
    linear_from_box_inequalities,
    split_dual_linear_box,
)
from ..exceptions import ProblemError
from ..problem import Problem
from ..solution import Solution
from ..solve_unconstrained import solve_unconstrained


def clarabel_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    r"""Solve a quadratic program using Clarabel.rs.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Notes
    -----
    Keyword arguments are forwarded as options to Clarabel.rs. For instance, we
    can call ``clarabel_solve_qp(P, q, G, h, u, tol_feas=1e-6)``. Clarabel
    options include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``max_iter``
         - Maximum number of iterations.
       * - ``time_limit``
         - Time limit for solve run in seconds (can be fractional).
       * - ``tol_gap_abs``
         - absolute duality-gap tolerance
       * - ``tol_gap_rel``
         - relative duality-gap tolerance
       * - ``tol_feas``
         - feasibility check tolerance (primal and dual)

    Check out the `API reference
    <https://oxfordcontrol.github.io/ClarabelDocs/stable/api_settings/#Clarabel.Settings>`_
    for details.

    Lower values for absolute or relative tolerances yield more precise
    solutions at the cost of computation time. See *e.g.* [Caron2022]_ for a
    primer on solver tolerances and residuals.
    """
    if initvals is not None and verbose:
        warnings.warn("Clarabel: warm-start values are ignored")
    P, q, G, h, A, b, lb, ub = problem.unpack()
    P, G, A = ensure_sparse_matrices("clarabel", P, G, A)
    if lb is not None or ub is not None:
        G, h = linear_from_box_inequalities(G, h, lb, ub, use_sparse=True)

    cones = []
    A_list = []
    b_list = []
    if A is not None and b is not None:
        A_list.append(A)
        b_list.append(b)
        cones.append(clarabel.ZeroConeT(b.shape[0]))
    if G is not None and h is not None:
        A_list.append(G)
        b_list.append(h)
        cones.append(clarabel.NonnegativeConeT(h.shape[0]))
    if not A_list:
        return solve_unconstrained(problem)

    settings = clarabel.DefaultSettings()
    settings.verbose = verbose
    for key, value in kwargs.items():
        setattr(settings, key, value)

    A_stack = spa.vstack(A_list, format="csc")
    b_stack = np.concatenate(b_list)
    try:
        solver = clarabel.DefaultSolver(
            P, q, A_stack, b_stack, cones, settings
        )
    except BaseException as exn:
        # The one we want to catch is a pyo3_runtime.PanicException
        # But see https://github.com/PyO3/pyo3/issues/2880
        raise ProblemError("Solver failed to build problem") from exn

    result = solver.solve()

    solution = Solution(problem)
    solution.obj = result.obj_val
    solution.extras = {
        "s": result.s,
        "status": result.status,
        "solve_time": result.solve_time,
    }

    solution.found = result.status == clarabel.SolverStatus.Solved
    if not solution.found:
        warnings.warn(f"Clarabel.rs terminated with status {result.status}")

    solution.x = np.array(result.x)
    meq = A.shape[0] if A is not None else 0
    solution.y = result.z[:meq] if meq > 0 else np.empty((0,))
    if G is not None:
        z, z_box = split_dual_linear_box(np.array(result.z[meq:]), lb, ub)
        solution.z = z
        solution.z_box = z_box
    else:  # G is None
        solution.z = np.empty((0,))
        solution.z_box = np.empty((0,))
    return solution


def clarabel_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using Clarabel.rs.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `Clarabel.rs`_.

    Parameters
    ----------
    P :
        Symmetric cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality matrix.
    h :
        Linear inequality vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = clarabel_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2024 Stéphane Caron and the qpsolvers contributors

import re
import sys
from os.path import abspath, dirname, join

sys.path.insert(0, abspath(".."))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx-mathjax-offline",
    "sphinx.ext.napoleon",  # before sphinx_autodoc_typehints
    "sphinx_autodoc_typehints",
]

# List of modules to be mocked up
autodoc_mock_imports = [
    "ecos",
    "gurobipy",
    "hpipm_python",
    "mosek",
    "nppro",
    "osqp",
    "qpoases",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "qpsolvers"
copyright = "2016-2024 Stéphane Caron and the qpsolvers contributors"
author = "Stéphane Caron"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The short X.Y version.
version = None

# The full version, including alpha/beta/rc tags.
release = None

# Read version info directly from the module's __init__.py
init_path = join(dirname(dirname(str(abspath(__file__)))), "qpsolvers")
with open(f"{init_path}/__init__.py", "r") as fh:
    for line in fh:
        match = re.match('__version__ = "((\\d.\\d).\\d)[a-z0-9\\-]*".*', line)
        if match is not None:
            release = match.group(1)
            version = match.group(2)
            break

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Override Pygments style.
pygments_style = "sphinx"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {}

# Output file base name for HTML help builder.
htmlhelp_basename = "qpsolversdoc"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solve quadratic programs."""

from typing import Optional, Union

import numpy as np
import scipy.sparse as spa

from .exceptions import NoSolverSelected
from .problem import Problem
from .solve_problem import solve_problem
from .solvers import available_solvers


def solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    using the QP solver selected by the ``solver`` keyword argument.

    Parameters
    ----------
    P :
        Symmetric cost matrix (most solvers require it to be definite as well).
    q :
        Cost vector.
    G :
        Linear inequality matrix.
    h :
        Linear inequality vector.
    A :
        Linear equality matrix.
    b :
        Linear equality vector.
    lb :
        Lower bound constraint vector. Can contain ``-np.inf``.
    ub :
        Upper bound constraint vector. Can contain ``+np.inf``.
    solver :
        Name of the QP solver, to choose in
        :data:`qpsolvers.available_solvers`. This argument is mandatory.
    initvals :
        Primal candidate vector :math:`x` values used to warm-start the solver.
    verbose :
        Set to ``True`` to print out extra information.

    Note
    ----
    In quadratic programming, the matrix :math:`P` should be symmetric. Many
    solvers (including CVXOPT, OSQP and quadprog) leverage this property and
    may return unintended results when it is not the case. You can set
    project :math:`P` on its symmetric part by:

    .. code:: python

        P = 0.5 * (P + P.transpose())

    Some solvers (like quadprog) will further require that :math:`P` is
    definite, while other solvers (like ProxQP or QPALM) can work with
    semi-definite matrices.

    Returns
    -------
    :
        Optimal solution if found, otherwise ``None``.

    Raises
    ------
    NoSolverSelected
        If the ``solver`` keyword argument is not set.

    ParamError
        If any solver parameter is incorrect.

    ProblemError
        If the problem is not correctly defined. For instance, if the solver
        requires a definite cost matrix but the provided matrix :math:`P` is
        not.

    SolverError
        If the solver failed during its execution.

    SolverNotFound
        If the requested solver is not in :data:`qpsolvers.available_solvers`.

    Notes
    -----
    Extra keyword arguments given to this function are forwarded to the
    underlying solver. For example, we can call OSQP with a custom absolute
    feasibility tolerance by ``solve_qp(P, q, G, h, solver='osqp',
    eps_abs=1e-6)``. See the :ref:`Supported solvers <Supported solvers>` page
    for details on the parameters available to each solver.

    There is no guarantee that a ``ValueError`` is raised if the provided
    problem is non-convex, as some solvers don't check for this. Rather, if the
    problem is non-convex and the solver fails because of that, then a
    ``ValueError`` will be raised.
    """
    if solver is None:
        raise NoSolverSelected(
            "Set the `solver` keyword argument to one of the "
            f"available solvers in {available_solvers}"
        )
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = solve_problem(problem, solver, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `SIP`_.

.. _SIP: https://github.com/joaospinto/sip_python

SIP is a general NLP solver based. It is based on the barrier augmented
Lagrangian method, which combines the interior point and augmented Lagrangian
methods. If you are using SIP in a scientific work, consider citing the
corresponding GitHub repository (or paper, if one has been released).
"""

import warnings
from typing import Optional, Union

import numpy as np
import scipy.sparse as spa
import sip_python as sip

from ..conversions import linear_from_box_inequalities, split_dual_linear_box
from ..exceptions import ProblemError
from ..problem import Problem
from ..solution import Solution

_cvxopt_available = True
try:
    from cvxopt import amd, spmatrix  # noqa: F401
except ImportError:
    _cvxopt_available = False


def sip_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    allow_non_psd_P: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using SIP.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP returned by the solver.

    Notes
    -----
    All other keyword arguments are forwarded as options to SIP. For
    instance, you can call ``sip_solve_qp(P, q, G, h, eps_abs=1e-6)``.
    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - max_iterations
         - The maximum number of iterations the solver can do.
       * - max_ls_iterations
         - The maximum cumulative number of line search iterations.
       * - min_iterations_for_convergence
         - The least number of iterations until we can declare convergence.
       * - num_iterative_refinement_steps
         - The number of iterative refinement steps.
       * - max_kkt_violation
         - The maximum allowed violation of the KKT system.
       * - max_merit_slope
         - The maximum allowed merit function slope.
       * - initial_regularization
         - The initial x-regularizatino to be applied on the LHS.
       * - regularization_decay_factor
         - The multiplicative decay of the x-regularization coefficient.
       * - tau
         - A parameter of the fraction-to-the-boundary rule.
       * - start_ls_with_alpha_s_max
         - Determines whether we start with alpha=alpha_s_max or alpha=1.
       * - initial_mu
         - The initial barrier function coefficient.
       * - mu_update_factor
         - Determines how much mu decreases per iteration.
       * - mu_min
         - The minimum barrier coefficient.
       * - initial_penalty_parameter
         - The initial penalty parameter of the Augmented Lagrangian.
       * - min_acceptable_constraint_violation_ratio
         - Least acceptable constraint violation ratio to not increase eta.
       * - penalty_parameter_increase_factor
         - By what factor to increase eta.
       * - penalty_parameter_decrease_factor
         - By what factor to decrease eta.
       * - max_penalty_parameter
         - The maximum allowed penalty parameter in the AL merit function.
       * - armijo_factor
         - Determines when we accept a line search step.
       * - line_search_factor
         - Determines how much to backtrack at each line search iteration.
       * - line_search_min_step_size
         - Determines when we declare a line search failure.
       * - min_merit_slope_to_skip_line_search
         - Min merit slope to skip the line search.
       * - dual_armijo_factor
         - Fraction of the primal merit decrease to allow on the dual update.
       * - min_allowed_merit_increase
         - The minimum allowed merit function increase in the dual update.
       * - enable_elastics
         - Whether to enable the usage of elastic variables.
       * - elastic_var_cost_coeff
         - Elastic variables cost coefficient.
       * - enable_line_search_failures
         - Halts the optimization process if a good step is not found.
       * - print_logs
         - Determines whether we should print the solver logs.
       * - print_line_search_logs
         - Determines whether we should print the line search logs.
       * - print_search_direction_logs
         - Whether we should print the search direction computation logs.
       * - print_derivative_check_logs
         - Whether to print derivative check logs when something looks off.
       * - only_check_search_direction_slope
         - Only derivative-check the search direction.
       * - assert_checks_pass
         - Handle checks with assert calls.

    This list may not be exhaustive.
    Check the `Settings` struct in the `solver code
    <https://github.com/joaospinto/sip/blob/main/sip/types.hpp>`__ for details.
    """
    P, q, G, h, A, b, lb, ub = problem.unpack()
    if lb is not None or ub is not None:
        G, h = linear_from_box_inequalities(
            G, h, lb, ub, use_sparse=problem.has_sparse
        )
    n: int = q.shape[0]

    # SIP does not support A, b, G, and h to be None.
    G = G if G is not None else spa.csr_matrix(np.zeros((0, n)))
    A = A if A is not None else spa.csr_matrix(np.zeros((0, n)))
    h = np.zeros((0,)) if h is None else h
    b = np.zeros((0,)) if b is None else b

    # Remove any infs from h.
    G[np.isinf(h), :] = 0.0
    h[np.isinf(h)] = 1.0

    if not isinstance(P, spa.csr_matrix):
        P = spa.csc_matrix(P)
        if verbose:
            warnings.warn("Converted P to a csc_matrix.")
    if not isinstance(G, spa.csr_matrix):
        G = spa.csr_matrix(G)
        if verbose:
            warnings.warn("Converted G to a csr_matrix.")
    if not isinstance(A, spa.csr_matrix):
        A = spa.csr_matrix(A)
        if verbose:
            warnings.warn("Converted A to a csr_matrix.")

    P.eliminate_zeros()
    G.eliminate_zeros()
    A.eliminate_zeros()

    P_T = spa.csc_matrix(P.T)
    if (
        (P.indices != P_T.indices).any()
        or (P.indptr != P_T.indptr).any()
        or (P.data != P_T.data).any()
    ):
        raise ProblemError("P should be symmetric.")

    if G is None and h is not None:
        raise ProblemError(
            "Inconsistent inequalities: G is not set but h is set"
        )
    if G is not None and h is None:
        raise ProblemError("Inconsistent inequalities: G is set but h is None")
    if A is None and b is not None:
        raise ProblemError(
            "Inconsistent inequalities: A is not set but b is set"
        )
    if A is not None and b is None:
        raise ProblemError("Inconsistent inequalities: A is set but b is None")

    k = None
    if allow_non_psd_P:
        eigenvalues, _eigenvectors = spa.linalg.eigsh(P, k=1, which="SM")
        k = -min(eigenvalues[0], 0.0) + 1e-3
    else:
        k = 1e-6

    # hess_L = P + k * spa.eye(n);
    # the code below avoids potential index cancellations.
    hess_L = spa.coo_matrix(P)
    upp_hess_L_rows = np.concatenate([hess_L.row, np.arange(n)])
    upp_hess_L_cols = np.concatenate([hess_L.col, np.arange(n)])
    upp_hess_L_data = np.concatenate([hess_L.data, k * np.ones(n)])
    hess_L = spa.coo_matrix(
        (upp_hess_L_data, (upp_hess_L_rows, upp_hess_L_cols)), shape=P.shape
    )
    hess_L.sum_duplicates()
    upp_hess_L = spa.triu(hess_L.tocsc())

    qs = sip.QDLDLSettings()
    qs.permute_kkt_system = True
    qs.kkt_pinv = sip.get_kkt_perm_inv(
        P=hess_L,
        A=A,
        G=G,
    )

    pd = sip.ProblemDimensions()
    pd.x_dim = n
    pd.s_dim = h.shape[0]
    pd.y_dim = b.shape[0]
    pd.upper_hessian_lagrangian_nnz = upp_hess_L.nnz
    pd.jacobian_c_nnz = A.nnz
    pd.jacobian_g_nnz = G.nnz
    pd.kkt_nnz, pd.kkt_L_nnz = sip.get_kkt_and_L_nnzs(
        P=hess_L,
        A=A,
        G=G,
        perm_inv=qs.kkt_pinv,
    )
    pd.is_jacobian_c_transposed = True
    pd.is_jacobian_g_transposed = True

    vars = sip.Variables(pd)

    if initvals is not None:
        vars.x[:] = initvals
    else:
        vars.x[:] = 0.0

    vars.s[:] = 1.0
    vars.y[:] = 0.0
    vars.e[:] = 0.0
    vars.z[:] = 1.0

    ss = sip.Settings()
    ss.max_iterations = 100
    ss.max_ls_iterations = 1000
    ss.max_kkt_violation = 1e-8
    ss.max_merit_slope = 1e-16
    ss.penalty_parameter_increase_factor = 2.0
    ss.mu_update_factor = 0.5
    ss.mu_min = 1e-16
    ss.max_penalty_parameter = 1e16
    ss.assert_checks_pass = True

    ss.print_logs = verbose
    ss.print_line_search_logs = verbose
    ss.print_search_direction_logs = verbose
    ss.print_derivative_check_logs = False

    for key, value in kwargs.items():
        try:
            setattr(ss, key, value)
        except AttributeError:
            if verbose:
                warnings.warn(
                    f"Received an undefined solver setting {key}\
                    with value {value}"
                )

    def mc(mci: sip.ModelCallbackInput) -> sip.ModelCallbackOutput:
        mco = sip.ModelCallbackOutput()

        Px = P.T @ mci.x

        mco.f = 0.5 * np.dot(Px, mci.x) + np.dot(q, mci.x)
        mco.c = A @ mci.x - b
        mco.g = G @ mci.x - h

        mco.gradient_f = Px + q
        mco.jacobian_c = A
        mco.jacobian_g = G
        mco.upper_hessian_lagrangian = upp_hess_L

        return mco

    solver = sip.Solver(ss, qs, pd, mc)

    output = solver.solve(vars)

    solution = Solution(problem)
    solution.extras = {"sip_output": output, "sip_vars": vars}
    solution.found = output.exit_status == sip.Status.SOLVED
    solution.obj = 0.5 * np.dot(P.T @ vars.x, vars.x) + np.dot(q, vars.x)
    solution.x = vars.x
    solution.y = vars.y
    if h is not None and vars.z is not None:
        z_sip = np.array(vars.z)
        z, z_box = split_dual_linear_box(z_sip, lb, ub)
        solution.z = z
        solution.z_box = z_box
    return solution


def sip_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    allow_non_psd_P: bool = False,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using SIP.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
        \underset{\mbox{minimize}}{x} &
            \frac{1}{2} x^T P x + q^T x \\
        \mbox{subject to}
            & G x \leq h                \\
            & A x = b                   \\
            & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `SIP
    <https://github.com/joaospinto/sip>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    verbose :
        Set to `True` to print out extra information.
    initvals :
        Warm-start guess vector. Not used.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = sip_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test a random QP solver on a constrained linear regression problem.

This example originates from:
    https://stackoverflow.com/a/74422084

See also:
    https://scaron.info/blog/simple-linear-regression-with-online-updates.html
"""

import random

import numpy as np

import qpsolvers
from qpsolvers import solve_ls

a = np.array([1.2, 2.3, 4.2])
b = np.array([1.0, 5.0, 6.0])
c = np.array([5.4, 6.2, 1.9])
m = np.vstack([a, b, c])
y = np.array([5.3, 0.9, 5.6])

# Objective: || [a b c] x - y ||^2
R = m.T
s = y

# Constraint: sum(x) = 1
A = np.ones((1, 3))
b = np.array([1.0])

# Constraint: x >= 0
lb = np.zeros(3)

if __name__ == "__main__":
    solver = random.choice(qpsolvers.available_solvers)
    x = solve_ls(R, s, A=A, b=b, lb=lb, solver=solver)
    print(f"Found solution {x=}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Output from a QP solver."""

from dataclasses import dataclass, field
from typing import Optional

import numpy as np

from .problem import Problem


@dataclass(frozen=False)
class Solution:
    """Solution returned by a QP solver for a given problem.

    Attributes
    ----------
    extras :
        Other outputs, specific to each solver.

    found :
        True if the solution was found successfully by a solver, False if the
        solver did not find a solution or detected an unfeasible problem,
        ``None`` if no solver was run.

    problem :
        Quadratic program the solution corresponds to.

    obj :
        Value of the primal objective at the solution (``None`` if no solution
        was found).

    x :
        Solution vector for the primal quadratic program (``None`` if no
        solution was found).

    y :
        Dual multipliers for equality constraints (``None`` if no solution was
        found, or if there is no equality constraint). The dimension of
        :math:`y` is equal to the number of equality constraints. The values
        :math:`y_i` can be either positive or negative.

    z :
        Dual multipliers for linear inequality constraints (``None`` if no
        solution was found, or if there is no inequality constraint). The
        dimension of :math:`z` is equal to the number of inequalities. The
        value :math:`z_i` for inequality :math:`i` is always positive.

        - If :math:`z_i > 0`, the inequality is active at the solution:
          :math:`G_i x = h_i`.
        - If :math:`z_i = 0`, the inequality is inactive at the solution:
          :math:`G_i x < h_i`.

    z_box :
        Dual multipliers for box inequality constraints (``None`` if no
        solution was found, or if there is no box inequality). The sign of
        :math:`z_{box,i}` depends on the active bound:

        - If :math:`z_{box,i} < 0`, then the lower bound :math:`lb_i = x_i` is
          active at the solution.
        - If :math:`z_{box,i} = 0`, then neither the lower nor the upper bound
          are active and :math:`lb_i < x_i < ub_i`.
        - If :math:`z_{box,i} > 0`, then the upper bound :math:`x_i = ub_i` is
          active at the solution.
    """

    problem: Problem
    extras: dict = field(default_factory=dict)
    found: Optional[bool] = None
    obj: Optional[float] = None
    x: Optional[np.ndarray] = None
    y: Optional[np.ndarray] = None
    z: Optional[np.ndarray] = None
    z_box: Optional[np.ndarray] = None

    def is_optimal(self, eps_abs: float) -> bool:
        """Check whether the solution is indeed optimal.

        Parameters
        ----------
        eps_abs :
            Absolute tolerance for the primal residual, dual residual and
            duality gap.

        Notes
        -----
        See for instance [Caron2022]_ for an overview of optimality conditions
        in quadratic programming.
        """
        return (
            self.primal_residual() < eps_abs
            and self.dual_residual() < eps_abs
            and self.duality_gap() < eps_abs
        )

    def primal_residual(self) -> float:
        r"""Compute the primal residual of the solution.

        The primal residual is:

        .. math::

            r_p := \max(\| A x - b \|_\infty, [G x - h]^+,
            [lb - x]^+, [x - ub]^+)

        were :math:`v^- = \min(v, 0)` and :math:`v^+ = \max(v, 0)`.

        Returns
        -------
        :
            Primal residual if it is defined, ``np.inf`` otherwise.

        Notes
        -----
        See for instance [Caron2022]_ for an overview of optimality conditions
        and why this residual will be zero at the optimum.
        """
        _, _, G, h, A, b, lb, ub = self.problem.unpack()
        if not self.found or self.x is None:
            return np.inf
        x = self.x
        return max(
            [
                0.0,
                np.max(G.dot(x) - h) if G is not None else 0.0,
                np.max(np.abs(A.dot(x) - b)) if A is not None else 0.0,
                np.max(lb - x) if lb is not None else 0.0,
                np.max(x - ub) if ub is not None else 0.0,
            ]
        )

    def dual_residual(self) -> float:
        r"""Compute the dual residual of the solution.

        The dual residual is:

        .. math::

            r_d := \| P x + q + A^T y + G^T z + z_{box} \|_\infty

        Returns
        -------
        :
            Dual residual if it is defined, ``np.inf`` otherwise.

        Notes
        -----
        See for instance [Caron2022]_ for an overview of optimality conditions
        and why this residual will be zero at the optimum.
        """
        P, q, G, _, A, _, lb, ub = self.problem.unpack()
        if not self.found or self.x is None:
            return np.inf
        zeros = np.zeros(self.x.shape)
        Px = P.dot(self.x)

        ATy = zeros
        if A is not None:
            if self.y is None:
                return np.inf
            ATy = A.T.dot(self.y)

        GTz = zeros
        if G is not None:
            if self.z is None:
                return np.inf
            GTz = G.T.dot(self.z)

        z_box = zeros
        if lb is not None or ub is not None:
            if self.z_box is None:
                return np.inf
            z_box = self.z_box

        p = np.linalg.norm(Px + q + GTz + ATy + z_box, np.inf)
        return p  # type: ignore

    def duality_gap(self) -> float:
        r"""Compute the duality gap of the solution.

        The duality gap is:

        .. math::

            r_g := | x^T P x + q^T x + b^T y + h^T z +
            lb^T z_{box}^- + ub^T z_{box}^+ |

        were :math:`v^- = \min(v, 0)` and :math:`v^+ = \max(v, 0)`.

        Returns
        -------
        :
            Duality gap if it is defined, ``np.inf`` otherwise.

        Notes
        -----
        See for instance [Caron2022]_ for an overview of optimality conditions
        and why this gap will be zero at the optimum.
        """
        P, q, _, h, _, b, lb, ub = self.problem.unpack()
        if not self.found or self.x is None:
            return np.inf
        xPx = self.x.T.dot(P.dot(self.x))
        qx = q.dot(self.x)

        hz = 0.0
        if h is not None:
            if self.z is None:
                return np.inf
            hz = h.dot(self.z)

        by = 0.0
        if b is not None:
            if self.y is None:
                return np.inf
            by = b.dot(self.y)

        lb_z_box = 0.0
        ub_z_box = 0.0
        if self.z_box is not None:
            if lb is not None:
                finite = np.asarray(lb != -np.inf).nonzero()
                z_box_neg = np.minimum(self.z_box, 0.0)
                lb_z_box = lb[finite].dot(z_box_neg[finite])
            if ub is not None:
                finite = np.asarray(ub != np.inf).nonzero()
                z_box_pos = np.maximum(self.z_box, 0.0)
                ub_z_box = ub[finite].dot(z_box_pos[finite])
        return abs(xPx + qx + hz + by + lb_z_box + ub_z_box)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2023 Stéphane Caron and the qpsolvers contributors

"""Internal objects."""

from .solvers import available_solvers as supported_solvers
from .solvers import solve_function as supported_solve
from .unsupported import available_solvers as unsupported_solvers
from .unsupported import solve_function as unsupported_solve

available_solvers = supported_solvers + unsupported_solvers

solve_function = {**supported_solve, **unsupported_solve}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Utility functions."""

from typing import Union

import numpy as np
import scipy.sparse as spa


def print_matrix_vector(
    A: Union[np.ndarray, spa.csc_matrix],
    A_label: str,
    b: np.ndarray,
    b_label: str,
    column_width: int = 24,
) -> None:
    """Print a matrix and vector side by side to the terminal.

    Parameters
    ----------
    A :
        Union[np.ndarray, spa.csc_matrix] to print.
    A_label :
        Label for A.
    b :
        np.ndarray to print.
    b_label :
        Label for b.
    column_width :
        Number of characters for the matrix and vector text columns.
    """
    if isinstance(A, np.ndarray) and A.ndim == 1:
        A = A.reshape((1, A.shape[0]))
    if isinstance(A, spa.csc_matrix):
        A = A.toarray()
    if A.shape[0] == b.shape[0]:
        A_string = f"{A_label} =\n{A}"
        b_string = f"{b_label} =\n{b.reshape((A.shape[0], 1))}"
    elif A.shape[0] > b.shape[0]:
        m = b.shape[0]
        A_string = f"{A_label} =\n{A[:m]}"
        b_string = f"{b_label} =\n{b.reshape(m, 1)}"
        A_string += f"\n{A[m:]}"
        b_string += "\n " * (A.shape[0] - m)
    else:  # A.shape[0] < b.shape[0]
        n = A.shape[0]
        k = b.shape[0] - n
        A_string = f"{A_label} =\n{A}"
        b_string = f"{b_label} =\n{b[:n].reshape(n, 1)}"
        A_string += "\n " * k
        b_string += f"\n{b[n:].reshape(k, 1)}"
    A_lines = A_string.splitlines()
    b_lines = b_string.splitlines()
    for i, A_line in enumerate(A_lines):
        print(A_line.ljust(column_width) + b_lines[i].ljust(column_width))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Functions to convert vector bounds into linear inequality constraints."""

from typing import Optional, Tuple, Union

import numpy as np
import scipy.sparse as spa

from ..exceptions import ProblemError


def concatenate_bound(
    G: Optional[Union[np.ndarray, spa.csc_matrix]],
    h: Optional[np.ndarray],
    b: np.ndarray,
    sign: float,
    use_sparse: bool,
) -> Tuple[Optional[Union[np.ndarray, spa.csc_matrix]], Optional[np.ndarray]]:
    """Append bound constraint vectors to inequality constraints.

    Parameters
    ----------
    G :
        Linear inequality matrix.
    h :
        Linear inequality vector.
    b :
        Bound constraint vector.
    sign :
        Sign factor: -1.0 for a lower and +1.0 for an upper bound.
    use_sparse :
        Use sparse matrices if true, dense matrices otherwise.

    Returns
    -------
    G : numpy.ndarray, scipy.sparse.csc_matrix, or None
        Updated linear inequality matrix.
    h : numpy.ndarray or None
        Updated linear inequality vector.
    """
    n = len(b)  # == number of optimization variables
    if G is None or h is None:
        G = sign * (spa.eye(n, format="csc") if use_sparse else np.eye(n))
        h = sign * b
    else:  # G is not None and h is not None
        if isinstance(G, np.ndarray):
            G = np.concatenate((G, sign * np.eye(n)), 0)
        elif isinstance(G, (spa.csc_matrix, spa.dia_matrix)):
            G = spa.vstack([G, sign * spa.eye(n)], format="csc")
        else:  # G is not an instance of a type we know
            name = type(G).__name__
            raise ProblemError(
                f"invalid type '{name}' for inequality matrix G"
            )
        h = np.concatenate((h, sign * b))
    return (G, h)


def linear_from_box_inequalities(
    G: Optional[Union[np.ndarray, spa.csc_matrix]],
    h: Optional[np.ndarray],
    lb: Optional[np.ndarray],
    ub: Optional[np.ndarray],
    use_sparse: bool,
) -> Tuple[Optional[Union[np.ndarray, spa.csc_matrix]], Optional[np.ndarray]]:
    """Append lower or upper bound vectors to inequality constraints.

    Parameters
    ----------
    G :
        Linear inequality matrix.
    h :
        Linear inequality vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    use_sparse :
        Use sparse matrices if true, dense matrices otherwise.

    Returns
    -------
    G : np.ndarray, spa.csc_matrix or None
        Updated linear inequality matrix.
    h : np.ndarray or None
        Updated linear inequality vector.
    """
    if lb is not None:
        G, h = concatenate_bound(G, h, lb, -1.0, use_sparse)
    if ub is not None:
        G, h = concatenate_bound(G, h, ub, +1.0, use_sparse)
    return (G, h)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for Problem class."""

import tempfile
import unittest

import numpy as np
import scipy.sparse as spa

from qpsolvers import ActiveSet, Problem, ProblemError

from .problems import get_sd3310_problem


class TestProblem(unittest.TestCase):
    """Test fixture for problems."""

    def setUp(self):
        self.problem = get_sd3310_problem()

    def test_unpack(self):
        P, q, G, h, A, b, lb, ub = self.problem.unpack()
        self.assertEqual(P.shape, self.problem.P.shape)
        self.assertEqual(q.shape, self.problem.q.shape)
        self.assertEqual(G.shape, self.problem.G.shape)
        self.assertEqual(h.shape, self.problem.h.shape)
        self.assertEqual(A.shape, self.problem.A.shape)
        self.assertEqual(b.shape, self.problem.b.shape)
        self.assertIsNone(lb)
        self.assertIsNone(ub)

    def test_check_inequality_constraints(self):
        problem = get_sd3310_problem()
        P, q, G, h, A, b, _, _ = problem.unpack()
        with self.assertRaises(ProblemError):
            Problem(P, q, G, None, A, b).check_constraints()
        with self.assertRaises(ProblemError):
            Problem(P, q, None, h, A, b).check_constraints()

    def test_check_equality_constraints(self):
        problem = get_sd3310_problem()
        P, q, G, h, A, b, _, _ = problem.unpack()
        with self.assertRaises(ProblemError):
            Problem(P, q, G, h, A, None).check_constraints()
        with self.assertRaises(ProblemError):
            Problem(P, q, G, h, None, b).check_constraints()

    def test_cond(self):
        active_set = ActiveSet(
            G_indices=range(self.problem.G.shape[0]),
            lb_indices=[],
            ub_indices=[],
        )
        self.assertGreater(self.problem.cond(active_set), 200.0)

    def test_cond_unconstrained(self):
        unconstrained = Problem(self.problem.P, self.problem.q)
        active_set = ActiveSet()
        self.assertAlmostEqual(
            unconstrained.cond(active_set), 124.257, places=4
        )

    def test_cond_no_equality(self):
        no_equality = Problem(
            self.problem.P, self.problem.q, self.problem.G, self.problem.h
        )
        active_set = ActiveSet(G_indices=range(self.problem.G.shape[0]))
        self.assertGreater(no_equality.cond(active_set), 200.0)

    def test_cond_sparse(self):
        sparse = Problem(spa.csc_matrix(self.problem.P), self.problem.q)
        active_set = ActiveSet()
        with self.assertRaises(ProblemError):
            sparse.cond(active_set)

    def test_check_matrix_shapes(self):
        Problem(np.eye(1), np.ones(1))
        Problem(np.array([1.0]), np.ones(1))

    def test_check_vector_shapes(self):
        Problem(np.eye(3), np.ones(shape=(3, 1)))
        Problem(np.eye(3), np.ones(shape=(1, 3)))
        Problem(np.eye(3), np.ones(shape=(3,)))
        with self.assertRaises(ProblemError):
            Problem(np.eye(3), np.ones(shape=(3, 2)))
        with self.assertRaises(ProblemError):
            Problem(np.eye(3), np.ones(shape=(3, 1, 1)))
        with self.assertRaises(ProblemError):
            Problem(np.eye(3), np.ones(shape=(1, 3, 1)))

    def test_save_load(self):
        problem = Problem(np.eye(3), np.ones(shape=(3, 1)))
        save_path = tempfile.mktemp() + ".npz"
        problem.save(save_path)
        reloaded = Problem.load(save_path)
        self.assertTrue(np.allclose(reloaded.P, problem.P))
        self.assertTrue(np.allclose(reloaded.q, problem.q))
        self.assertIsNone(reloaded.G)
        self.assertIsNone(reloaded.h)
        self.assertIsNone(reloaded.A)
        self.assertIsNone(reloaded.b)
        self.assertIsNone(reloaded.lb)
        self.assertIsNone(reloaded.ub)
